#include <deque>
#include <iostream>
#include <memory>
#include <mutex>
#include <ostream>
#include <sstream>
#include <string>
#include <thread>

#include "semaphore.h"

template <typename T, typename A = std::allocator<T>>
class sync_queue
{
public:

    void push_back(T&& item)
    {
        queue_.push_back(item);
    }

    // This function combines deque::front() and deque::pop_front(),
    // i.e. it returns the front item *and* removes it from the queue!
    T pop_front()
    {
        T item;

        item = queue_.front();
        queue_.pop_front();

        return item;
    }

private:

    std::deque<T, A> queue_;

};

sync_queue<std::string> the_queue;

void producer_func(int index)
{
    for (;;)
    {
        std::stringstream stream;
        stream << "Message from producer " << index;
        the_queue.push_back(stream.str());
    }
}

void consumer_func()
{
    for (;;)
    {
        std::string str = the_queue.pop_front();
        std::cout << str << '\n';
    }
}


int main()
{
    std::thread producer1(producer_func, 1);
    std::thread producer2(producer_func, 2);
    std::thread producer3(producer_func, 3);
    std::thread producer4(producer_func, 4);
    std::thread producer5(producer_func, 5);
    std::thread producer6(producer_func, 6);
    std::thread producer7(producer_func, 7);
    std::thread producer8(producer_func, 8);
    std::thread consumer(consumer_func);

    consumer.join();
    producer8.join();
    producer7.join();
    producer6.join();
    producer5.join();
    producer4.join();
    producer3.join();
    producer2.join();
    producer1.join();
}
